/** @file drv_wrapper.cpp
 *
 *  @brief This file contains driver ioctl interface related code.
 *
 *  (C) Copyright 2008-2012 Marvell International Ltd. All Rights Reserved
 *
 *  MARVELL CONFIDENTIAL
 *  The source code contained or described herein and all documents related to
 *  the source code ("Material") are owned by Marvell International Ltd or its
 *  suppliers or licensors. Title to the Material remains with Marvell International Ltd
 *  or its suppliers and licensors. The Material contains trade secrets and
 *  proprietary and confidential information of Marvell or its suppliers and
 *  licensors. The Material is protected by worldwide copyright and trade secret
 *  laws and treaty provisions. No part of the Material may be used, copied,
 *  reproduced, modified, published, uploaded, posted, transmitted, distributed,
 *  or disclosed in any way without Marvell's prior express written permission.
 *
 *  No license under any patent, copyright, trade secret or other intellectual
 *  property right is granted to or conferred upon you by disclosure or delivery
 *  of the Materials, either expressly, by implication, inducement, estoppel or
 *  otherwise. Any license under such intellectual property rights must be
 *  express and approved by Marvell in writing.
 *
 */

/** @file drv_wrapper.cpp
 *
 *  @brief This file contains driver ioctl interface related code.
 *
 *  Copyright (C) 2009-2010, Marvell International Ltd.
 *  All Rights Reserved
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <stdlib.h>
#include <poll.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <linux/if.h>
#include <linux/ioctl.h>
#include <linux/wireless.h>
#ifdef NONPLUG_SUPPORT
#ifndef MARVELL_BT_STACK
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#endif
#endif
#include "drv_wrapper.h"

#ifdef MFG_UPDATE

#include "mfgbridge.h"
#include "mfgdebug.h"

#endif

/********************************************************
        Local Variables
********************************************************/
#define IW_MAX_PRIV_DEF   128
#define MLAN_ETH_PRIV               (SIOCDEVPRIVATE + 14)

#define BRDG_OFFSET 4

static int ethio_flag = 0;

/** Private command structure */
typedef struct eth_priv_cmd
{
    /** Command buffer */
    char *buf;
    /** Used length */
    unsigned int used_len;
    /** Total length */
    unsigned int total_len;
} eth_cmd;

static struct ifreq userdata;
static struct iw_priv_args priv_args[IW_MAX_PRIV_DEF];
static int sockfd = 0, ioctl_val, subioctl_val;
#ifdef NONPLUG_SUPPORT
static int hci_sock = 0;
static char *hci_addr = NULL;
static char addr[18];
static char *hci_intf = "hci0";
#endif

#ifdef MFG_UPDATE
static struct _drv_cb Driver;
#endif // MFG_UPDATE
/********************************************************
        Local Functions
********************************************************/
/**
 *  @brief Get private info.
 *
 *  @param ifname   A pointer to net name
 *  @return         Number of private IOCTLs, -1 on failure
 */
static int
get_private_info(const char *ifname)
{
    struct iwreq iwr;
    int ret = 0;
    struct iw_priv_args *pPriv = priv_args;

    memset(&iwr, 0, sizeof(iwr));
    strncpy(iwr.ifr_name, ifname, IFNAMSIZ);
    iwr.u.data.pointer = (caddr_t) pPriv;
    iwr.u.data.length = IW_MAX_PRIV_DEF;
    iwr.u.data.flags = 0;

    if (ioctl(sockfd, SIOCGIWPRIV, &iwr) < 0) {
        perror("ioctl[SIOCGIWPRIV]");
        ret = -1;
    } else {
        /* Return the number of private ioctls */
        ret = iwr.u.data.length;
    }

    return ret;
}

/**
 *  @brief Get Sub command ioctl number
 *
 *  @param i            command index
 *  @param priv_cnt     Total number of private ioctls availabe in driver
 *  @param ioctl_val    A pointer to return ioctl number
 *  @param subioctl_val A pointer to return sub-ioctl number
 *  @return             0 on success, otherwise -1
 */
static int
get_subioctl_no(int i, int priv_cnt, int *ioctl_val, int *subioctl_val)
{
    int j;

    if (priv_args[i].cmd >= SIOCDEVPRIVATE) {
        *ioctl_val = priv_args[i].cmd;
        *subioctl_val = 0;
        return 0;
    }

    j = -1;
    while ((++j < priv_cnt)
           && ((priv_args[j].name[0] != '\0') ||
               (priv_args[j].set_args != priv_args[i].set_args) ||
               (priv_args[j].get_args != priv_args[i].get_args))) {
    }

    /* If not found... */
    if (j == priv_cnt) {
        printf("Invalid private ioctl definition for: 0x%x\n",
               priv_args[i].cmd);
        return -1;
    }

    /* Save ioctl numbers */
    *ioctl_val = priv_args[j].cmd;
    *subioctl_val = priv_args[i].cmd;

    return 0;
}

/**
 *  @brief Get ioctl number
 *
 *  @param ifname       Interface name
 *  @param priv_cmd     Private command name
 *  @param ioctl_val    A pointer to return ioctl number
 *  @param subioctl_val A pointer to return sub-ioctl number
 *  @return             0 on success, otherwise -1
 */
static int
get_ioctl_no(const char *ifname,
             const char *priv_cmd, int *ioctl_val, int *subioctl_val)
{
    int i, priv_cnt, ret = -1;

    priv_cnt = get_private_info(ifname);

    /* Are there any private ioctls? */
    if (priv_cnt <= 0 || priv_cnt > IW_MAX_PRIV_DEF) {
        /* Could skip this message ? */
        printf("%-8.8s  no private ioctls.\n", ifname);
        ethio_flag = 1;
        ret = 0;
    } else {
        for (i = 0; i < priv_cnt; i++) {
            if (priv_args[i].name[0]
                && !strcmp(priv_args[i].name, priv_cmd)) {
                ret = get_subioctl_no(i, priv_cnt, ioctl_val, subioctl_val);
                break;
            }
        }
    }
    return ret;
}

#ifdef NONPLUG_SUPPORT
static int
get_hci_dev_info(int s, int dev_id, long arg)
{
#ifndef MARVELL_BT_STACK
  struct hci_dev_info di = { dev_id:dev_id };

    if (ioctl(s, HCIGETDEVINFO, (void *) &di))
        return 0;

    if (!strcmp(di.name, hci_intf)) {
        ba2str(&di.bdaddr, addr);
        hci_addr = addr;
    }
#endif
    return 0;
}

/**
 *  @brief Get HCI driver info
 *
 *  @return  0 --- if already loadded, otherwise -1
 */
int
drv_wrapper_get_hci_info(char *cmdname)
{
    int ret = 0;

#ifndef MARVELL_BT_STACK
    if (!cmdname) {
        printf("Interface name is not present\n");
        return -1;
    }
    hci_intf = cmdname;
    hci_for_each_dev(HCI_UP, get_hci_dev_info, 0);

    ret = (hci_addr == NULL) ? -1 : 0;
#endif

    return ret;
}

/**
 *  @brief drv wrapper initialize HCI
 *
 *  @param intfname Interface name
 *  @return  0 on success, otherwise -1
 */
int
drv_wrapper_init_hci(char *cmdname)
{
    int ret = 0;

#ifdef MARVELL_BT_STACK
    char dev_name[20];

    strcpy(dev_name, "/dev/mbtchar0");
    hci_sock = open(dev_name, O_RDWR | O_NOCTTY);
    if (hci_sock >= 0) {
        printf("Use device %s\n", dev_name);
    } else {
        printf("Cannot open device %s", dev_name);
        ret = -1;
    }
#else
    if (!cmdname) {
        printf("Interface name is not present\n");
        return -1;
    }
    hci_intf = cmdname;
    hci_addr = NULL;
    printf("Initialize drvwrapper for BT ....\n");
    hci_for_each_dev(HCI_UP, get_hci_dev_info, 0);
    if (hci_addr == NULL) {
        printf("BT interface is not present\n");
        ret = -1;
    } else {
        int hci_dev_id;
        hci_dev_id = hci_get_route((bdaddr_t *) hci_addr);
        hci_sock = hci_open_dev(hci_dev_id);
        ret = 0;
    }
#endif
    return ret;
}

/**
 *  @brief drv wrapper de-initialize HCI
 *
 *  @return  0 on success, otherwise -1
 */
int
drv_wrapper_deinit_hci()
{
    int ret = 0;

    printf("De-Initialize drvwrapper for BT....\n");
    if (hci_sock)
        close(hci_sock);
    hci_addr = NULL;
    return ret;
}

/**
 *  @brief drv wrapper send HCI command
 *
 *  @return  0 on success, otherwise error code
 */
int
drv_wrapper_send_hci_command(short ogf, short ocf, unsigned char *in_buf,
                             int in_buf_len, unsigned char *out_buf,
                             int *out_buf_len)
{
#define EVENT_BUF_SIZE 400
    int status = 0;
#ifdef MARVELL_BT_STACK
    int len;
#else
    struct hci_filter flt;
#endif
    int dummy_buf[512];
    int avail = 0;

#ifdef MARVELL_BT_STACK
    len = write(hci_sock, in_buf, in_buf_len);
    if (len != in_buf_len) {
        printf("Failed to write %d bytes (written %d)\n", in_buf_len, len);
        status = -1;
    }
#else
    /* Setup filter */
    hci_filter_clear(&flt);
    hci_filter_set_ptype(HCI_EVENT_PKT, &flt);
    hci_filter_all_events(&flt);
    if (setsockopt(hci_sock, SOL_HCI, HCI_FILTER, &flt, sizeof(flt)) < 0) {
        printf("HCI filter setup failed\n");
        return (-1);
    }

    status =
        hci_send_cmd(hci_sock, ogf, ocf, in_buf_len - BRDG_OFFSET,
                     in_buf + BRDG_OFFSET);
#endif
    if (status < 0) {
        printf("Failed to send command (OGF = %04x OCF = %04x)\n", ogf, ocf);
        return status;
    }

    do {
        *out_buf_len = read(hci_sock, out_buf, EVENT_BUF_SIZE);
        printf("Rx Event %X for %02X %02X \n", out_buf[1], out_buf[4],
               out_buf[5]);
        printf("In Cmd %02X %02X \n", in_buf[1], in_buf[2]);
    } while ((out_buf[1] != 0xFF)
             && !(out_buf[1] != 0xF && out_buf[4] == in_buf[1]
                  && out_buf[5] == in_buf[2]));

    if (out_buf[1] == 0xFF)
        avail = read(hci_sock, dummy_buf, EVENT_BUF_SIZE);
    return status;
}
#endif /* NONPLUG_SUPPORT */

/**
 *  @brief Get driver info
 *
 *  @return  0 --- if already loadded, otherwise -1
 */
int
drv_wrapper_get_info(char *cmdname)
{
    struct ifreq ifr;
    int sk, ret = 0;

    /* Open socket to communicate with driver */
    sk = socket(AF_INET, SOCK_STREAM, 0);
    if (sk < 0) {
        printf("Failed to create socket\n");
        return -1;
    }
    strncpy(ifr.ifr_name, cmdname, IFNAMSIZ);
    if (ioctl(sk, SIOCGIFFLAGS, &ifr) < 0)
        ret = -1;
    close(sk);
    return ret;
}

/**
 *  @brief drv wrapper initialize
 *
 *  @return  0 on success, otherwise -1
 */
int
drv_wrapper_init(char *cmdname)
{
    int ret = 0;

    printf("Initialize drvwrapper ....\n");
    /* Open socket to communicate with driver */
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (sockfd < 0) {
        printf("Invalid sockfd, function return -1\n");
        return -1;
    }
    if (get_ioctl_no(cmdname, "hostcmd", &ioctl_val, &subioctl_val)
        == -1) {
        printf("No private IOCTL present for command \"hostcmd\"\n");
        close(sockfd);
        sockfd = 0;
        return -1;
    }

    return ret;
}

/**
 *  @brief drv wrapper de-initialize
 *
 *  @return  0 on success, otherwise -1
 */
int
drv_wrapper_deinit()
{
    int ret = 0;

    printf("De-Initialize drvwrapper ....\n");
    if (sockfd)
        close(sockfd);
    ioctl_val = 0;
    subioctl_val = 0;
    sockfd = 0;
    return ret;
}

#ifdef NONPLUG_SUPPORT

/*
 * int drv_wrapper_send_hci_command(short ogf, short ocf,  unsigned char *in_buf, int in_buf_len,
 * unsigned char *out_buf, int *out_buf_len)
 */
int
drv_proc_hci_command(drv_cb * drv, unsigned char *buf, int *msglen, int buflen)
{
#define EVENT_BUF_SIZE 400
    int status = 0;
#ifdef MARVELL_BT_STACK
    int len;
#else
    struct hci_filter flt;
#endif
    int dummy_buf[512];
    int avail = 0;

    unsigned short ogf;
    unsigned short ocf;
    unsigned char *evntbuf = drv->cb_buf;

#ifdef MARVELL_BT_STACK
    len = write(hci_sock, buf, *msglen);
    if (len != *msglen) {
        printf("Failed to write %d bytes (written %d)\n", *msglen, len);
        status = -1;
	ogf = buf[2] >> 2;
	ocf = ((buf[2] & 0x03) << 8) + buf[1];
    }
#else
    /* Setup filter */
    hci_filter_clear(&flt);
    hci_filter_set_ptype(HCI_EVENT_PKT, &flt);
    hci_filter_all_events(&flt);
    if (setsockopt(hci_sock, SOL_HCI, HCI_FILTER, &flt, sizeof(flt)) < 0) {
        mfg_dprint(DBG_ERROR, "HCI filter setup failed\n");
        return -1;
    }
    ogf = buf[2] >> 2;
    ocf = ((buf[2] & 0x03) << 8) + buf[1];

    status = hci_send_cmd(hci_sock, ogf, ocf, *msglen - 4, buf + 4);
#endif
    if (status < 0) {
        mfg_dprint(DBG_ERROR,
                   "Failed to send command (OGF = %04x OCF = %04x)\n",
                   ogf, ocf);
        return status;
    }

    memset(evntbuf, 0, MAXBUF);

    do {
        *msglen = read(hci_sock, evntbuf, EVENT_BUF_SIZE);
        printf("Rx Event %X for %2.2X %2.2X \n", evntbuf[1], evntbuf[4],
               evntbuf[5]);
        printf("In Cmd  %2.2X %2.2X \n", buf[1], buf[2]);

    } while ((evntbuf[1] != 0xFF)
             && !(evntbuf[1] != 0xF && evntbuf[4] == buf[1]
                  && evntbuf[5] == buf[2]));

    if (evntbuf[1] == 0xFF)
        avail = read(hci_sock, dummy_buf, EVENT_BUF_SIZE);

    memcpy(buf, evntbuf, *msglen);
    return status;
}

#endif /* NONPLUG_SUPPORT */

int
drv_config_default(drv_config * drv_conf)
{
    memset(drv_conf, 0, sizeof(drv_config));

#ifdef NONPLUG_SUPPORT
    sprintf(drv_conf->hci_port, "hci0");
#endif
    sprintf(drv_conf->wlan_port, "mlan0");

    return 0;
}

int
drv_init(struct _bridge_cb *bridge, drv_config * drv_conf)
{
    drv_cb *drv;

    bridge->drv = &Driver;
    drv = bridge->drv;

#ifdef NONPLUG_SUPPORT
    drv->hci_ifname = drv_conf->hci_port;
#endif

    drv->wlan_ifname = drv_conf->wlan_port;
    drv->load_script = drv_conf->load_script;
    drv->unload_script = drv_conf->unload_script;

    /** Initialize drvwrapper, if driver already loadded */
    /** Init WiFi */
    if (!drv_wrapper_get_info(drv->wlan_ifname))
        drv_wrapper_init(drv->wlan_ifname);
#ifdef NONPLUG_SUPPORT
    /** Init BT */
    if (!drv_wrapper_get_hci_info(drv->hci_ifname))
        drv_wrapper_init_hci(drv->hci_ifname);
#endif
    mfg_dprint(DBG_MINFO, "DRV:  driver is initialized.\n");
    return 0;
}

int
drv_proc_wlan_command(drv_cb * drv, unsigned char *buf, int *rsplen, int buflen)
{
    int status = -1;
    HostCmd_DS_Gen *hostcmd_hdr = (HostCmd_DS_Gen *) buf;
    char *ifname = drv->wlan_ifname;
    eth_cmd ether_cmd;

    int eth_header_len;
    char *buffer;

    if (ethio_flag == 1) {
        struct ifreq ifr;
        char *p_cmd = "MRVL_CMDhostcmd";
        eth_header_len = strlen(p_cmd) + BRDG_OFFSET;
        eth_cmd *cmd = &ether_cmd;

        buffer = (char *) malloc(buflen + eth_header_len);
        if (buffer == NULL) {
            printf("can't allocate buffer \n");
            exit(0);
        }

        strncpy((char *) buffer, p_cmd, strlen(p_cmd));
        memcpy(buffer + eth_header_len, buf, buflen);

        /* Fill up buffer */
        cmd->buf = buffer;
        cmd->used_len = 0;
        cmd->total_len = buflen + eth_header_len;

        /* Perform IOCTL */
        memset(&ifr, 0, sizeof(struct ifreq));
        strncpy(ifr.ifr_ifrn.ifrn_name, ifname, strlen(ifname));
        ifr.ifr_ifru.ifru_data = (void *) cmd;

        if (ioctl(sockfd, MLAN_ETH_PRIV, &ifr)) {
            printf("ether bridge: hostcmd fail\n");
        }

        *rsplen = hostcmd_hdr->size;
        memcpy(buf, cmd->buf + eth_header_len, *rsplen);
        free(buffer);
    } else {

        *rsplen = 0;
        strncpy(userdata.ifr_name, ifname, strlen(ifname));
        userdata.ifr_data = (char *) buf;

        mfg_dprint(DBG_GINFO, "DRV:  send host cmd thr ioctl\n");

        status = ioctl(sockfd, ioctl_val, &userdata);
        *rsplen = hostcmd_hdr->size;

    }
    mfg_dprint(DBG_GINFO, "DRV:  host cmd is completed\n");
    return status;
}

int
drv_load_driver(struct _drv_cb *drv, int drv_if)
{
    char *command = drv->command;
    char *script = drv->load_script;

    if (!script) {
        printf("Load script is not provided\n");
        return -1;
    }

    if (!drv_wrapper_get_info(drv->wlan_ifname)) {
        printf("Driver already loaded\n");
    } else {
        sprintf(command, "sh %s", script);
        printf("Load driver ......\n");
        if (system(command) != 0) {
            printf("Failed to run the script\n");
            return -1;
        } else {
            drv_wrapper_init(drv->wlan_ifname);
        }
    }
    return 0;
}

int
drv_unload_driver(struct _drv_cb *drv, int drv_if)
{

    int ret = -1;
    char *command = drv->command;
    char *script = drv->unload_script;

    if (!script) {
        printf("Unload script is not provided\n");
        return -1;
    }
    if (!drv_wrapper_get_info(drv->wlan_ifname)) {
        sprintf(command, "sh %s", script);
        drv_wrapper_deinit();
        printf("Unload driver ......\n");
        ret = system(command);
        if (ret) {
            printf("Failed to run the script\n");
            return -1;
        }
    } else {
        printf("No such device\n");
    }
    return 0;
}
