/** @file bt_main.c
  *  
  * @brief This file contains the major functions in BlueTooth
  * driver. It includes init, exit, open, close and main
  * thread etc..
  * 
  * Copyright (C) 2007-2012, Marvell International Ltd. 
  *   
  * This software file (the "File") is distributed by Marvell International 
  * Ltd. under the terms of the GNU General Public License Version 2, June 1991 
  * (the "License").  You may use, redistribute and/or modify this File in 
  * accordance with the terms and conditions of the License, a copy of which 
  * is available along with the File in the gpl.txt file or by writing to 
  * the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 
  * 02111-1307 or on the worldwide web at http://www.gnu.org/licenses/gpl.txt.
  *
  * THE FILE IS DISTRIBUTED AS-IS, WITHOUT WARRANTY OF ANY KIND, AND THE 
  * IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE 
  * ARE EXPRESSLY DISCLAIMED.  The License provides additional details about 
  * this warranty disclaimer.
  *
  */
/**
  * @mainpage M-BT Linux Driver
  *
  * @section overview_sec Overview
  *
  * The M-BT is a Linux reference driver for Marvell SDIO Bluetooth chipset.
  * 
  * @section copyright_sec Copyright
  *
  * Copyright (C) 2007-2011, Marvell International Ltd.
  *
  */
#include "include.h"
#include "mbt_char.h"

/** Version */
#define VERSION "M2614105"

/** Driver version */
char driver_version[] =
	"sd8797-%s-" VERSION
	"-(" "FP" FPNUM ")"
#ifdef DEBUG_LEVEL2
	"-dbg"
#endif
	" ";

/** Declare and initialize fw_version */
char fw_version[32] = "0.0.0.p0";

#define AID_SYSTEM        1000	/* system server */

#define AID_BLUETOOTH     1002	/* bluetooth subsystem */

/** Define module name */
#define MODULE_NAME  "bt_fm_nfc"

/** Define FM char device name length */
#define FM_CHAR_NAME_LEN   8
/** Define NFC char device name length */
#define NFC_CHAR_NAME_LEN  9

/** Declaration of chardev class */
struct class *chardev_class;

/** Interface specific variables */
int fmchar_minor = 0;
int nfcchar_minor = 0;

/** Default Driver mode */
int drv_mode = (DRV_MODE_BT | DRV_MODE_FM | DRV_MODE_NFC);

/** Firmware flag */
int fw = 1;
/** default powermode */
int psmode = 1;
/** Init config file (MAC address, register etc.) */
char *init_cfg = NULL;

/** Setting drvdbg value based on DEBUG level */
#ifdef	DEBUG_LEVEL1
#ifdef  DEBUG_LEVEL2
#define DEFAULT_DEBUG_MASK  (0xffffffff & ~DBG_EVENT)
#else
#define DEFAULT_DEBUG_MASK  (DBG_MSG | DBG_FATAL | DBG_ERROR)
#endif /* DEBUG_LEVEL2 */
u32 drvdbg = DEFAULT_DEBUG_MASK;
#endif

#ifdef SDIO_SUSPEND_RESUME
/** PM keep power */
int pm_keep_power = 1;
#endif



/** 
 *  @brief Alloc bt device 
 *  
 *  @return    pointer to structure mbt_dev or NULL
 */
struct mbt_dev *alloc_mbt_dev(void)
{
	struct mbt_dev *mbt_dev;
	ENTER();

	mbt_dev = kzalloc(sizeof(struct mbt_dev), GFP_KERNEL);
	if (!mbt_dev) {
		LEAVE();
		return NULL;
	}

	LEAVE();
	return mbt_dev;
}

/** 
 *  @brief Alloc fm device 
 *  
 *  @return    pointer to structure fm_dev or NULL
 */
struct fm_dev *alloc_fm_dev(void)
{
	struct fm_dev *fm_dev;
	ENTER();

	fm_dev = kzalloc(sizeof(struct fm_dev), GFP_KERNEL);
	if (!fm_dev) {
		LEAVE();
		return NULL;
	}

	LEAVE();
	return fm_dev;
}

/** 
 *  @brief Alloc nfc device 
 *  
 *  @return    pointer to structure nfc_dev or NULL
 */
struct nfc_dev *alloc_nfc_dev(void)
{
	struct nfc_dev *nfc_dev;
	ENTER();

	nfc_dev = kzalloc(sizeof(struct nfc_dev), GFP_KERNEL);
	if (!nfc_dev) {
		LEAVE();
		return NULL;
	}

	LEAVE();
	return nfc_dev;
}

/** 
 *  @brief Frees m_dev
 *  
 *  @return    N/A
 */
void free_m_dev(struct m_dev *m_dev)
{
	ENTER();
	if (m_dev->dev_pointer)
		kfree(m_dev->dev_pointer);
	LEAVE();
}

/** 
 *  @brief This function verify the received event pkt
 *
 *  Event format:
 *  +--------+--------+--------+--------+--------+
 *  | Event  | Length |  ncmd  |      Opcode     |
 *  +--------+--------+--------+--------+--------+
 *  | 1-byte | 1-byte | 1-byte |      2-byte     |
 *  +--------+--------+--------+--------+--------+
 *  
 *  @param priv    A pointer to bt_private structure
 *  @param skb     A pointer to rx skb 
 *  @return 	   BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
int check_evtpkt(bt_private * priv, struct sk_buff *skb)
{
	struct hci_event_hdr *hdr = (struct hci_event_hdr *) skb->data;
	struct hci_ev_cmd_complete *ec;
	u16 opcode, ocf;
	u8 ret = BT_STATUS_SUCCESS;
	ENTER();
	if (!priv->bt_dev.sendcmdflag) {
		ret = BT_STATUS_FAILURE;
		goto exit;
	}
	if (hdr->evt == HCI_EV_CMD_COMPLETE) {
		ec = (struct hci_ev_cmd_complete *) (skb->data + HCI_EVENT_HDR_SIZE);
		opcode = __le16_to_cpu(ec->opcode);
		ocf = hci_opcode_ocf(opcode);
		PRINTM(CMD, "BT: CMD_COMPLTE ocf=0x%x, send_cmd_ocf=0x%x\n", ocf,
			   priv->bt_dev.send_cmd_ocf);
		if (ocf != priv->bt_dev.send_cmd_ocf) {
			ret = BT_STATUS_FAILURE;
			goto exit;
		}
		switch (ocf) {
		case BT_CMD_MODULE_CFG_REQ:
		case BT_CMD_BLE_DEEP_SLEEP:
		case BT_CMD_CONFIG_MAC_ADDR:
		case BT_CMD_CSU_WRITE_REG:
		case BT_CMD_AUTO_SLEEP_MODE:
		case BT_CMD_HOST_SLEEP_CONFIG:
		case BT_CMD_SDIO_PULL_CFG_REQ:
			priv->bt_dev.sendcmdflag = FALSE;
			priv->adapter->cmd_complete = TRUE;
			wake_up_interruptible(&priv->adapter->cmd_wait_q);
			break;
		case BT_CMD_GET_FW_VERSION:
			{
			u8 *pos = (skb->data + HCI_EVENT_HDR_SIZE +
            	       sizeof(struct hci_ev_cmd_complete) + 1);
			snprintf(fw_version, sizeof(fw_version), "%u.%u.%u.p%u",
					 pos[2], pos[1], pos[0], pos[3]);
			priv->bt_dev.sendcmdflag = FALSE;
			priv->adapter->cmd_complete = TRUE;
			wake_up_interruptible(&priv->adapter->cmd_wait_q);
			break;
			}
		case BT_CMD_HOST_SLEEP_ENABLE:
			priv->bt_dev.sendcmdflag = FALSE;
			break;
		default:
			ret = BT_STATUS_FAILURE;
			break;
		}
	}
exit:
	if (ret == BT_STATUS_SUCCESS)
		kfree_skb(skb);
	LEAVE();
	return ret;
}

/** 
 *  @brief This function process the received event
 *
 *  Event format:
 *  +--------+--------+--------+--------+-----+
 *  |   EC   | Length |           Data        |
 *  +--------+--------+--------+--------+-----+
 *  | 1-byte | 1-byte |          n-byte       |
 *  +--------+--------+--------+--------+-----+
 *  
 *  @param priv    A pointer to bt_private structure
 *  @param skb     A pointer to rx skb 
 *  @return 	   BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
int bt_process_event(bt_private * priv, struct sk_buff *skb)
{
	u8 ret = BT_STATUS_SUCCESS;
	struct m_dev *m_dev = &(priv->bt_dev.m_dev[BT_SEQ]);
	BT_EVENT *pEvent;

	ENTER();
	pEvent = (BT_EVENT *) skb->data;
	if (pEvent->EC != 0xff) {
		PRINTM(CMD, "BT: Not Marvell Event=0x%x\n", pEvent->EC);
		ret = BT_STATUS_FAILURE;
		goto exit;
	}
	switch (pEvent->data[0]) {
	case BT_CMD_AUTO_SLEEP_MODE:
		if (pEvent->data[2] == BT_STATUS_SUCCESS) {
			if (pEvent->data[1] == BT_PS_ENABLE)
				priv->adapter->psmode = 1;
			else
				priv->adapter->psmode = 0;
			PRINTM(CMD, "BT: PS Mode %s:%s\n", m_dev->name,
				   (priv->adapter->psmode) ? "Enable" : "Disable");

		} else {
			PRINTM(CMD, "BT: PS Mode Command Fail %s\n", m_dev->name);
		}
		break;
	case BT_CMD_HOST_SLEEP_CONFIG:
		if (pEvent->data[3] == BT_STATUS_SUCCESS) {
			PRINTM(CMD, "BT: %s: gpio=0x%x, gap=0x%x\n", m_dev->name,
				   pEvent->data[1], pEvent->data[2]);
		} else {
			PRINTM(CMD, "BT: %s: HSCFG Command Fail\n", m_dev->name);
		}
		break;
	case BT_CMD_HOST_SLEEP_ENABLE:
		if (pEvent->data[1] == BT_STATUS_SUCCESS) {
			priv->adapter->hs_state = HS_ACTIVATED;
			if (priv->adapter->suspend_fail == FALSE) {
#ifdef SDIO_SUSPEND_RESUME
#ifdef MMC_PM_KEEP_POWER
#ifdef MMC_PM_FUNC_SUSPENDED
				bt_is_suspended(priv);
#endif
#endif
#endif
				wake_up_interruptible(&priv->adapter->cmd_wait_q);
			}
			if (priv->adapter->psmode)
				priv->adapter->ps_state = PS_SLEEP;
			PRINTM(CMD, "BT: EVENT %s: HS ACTIVATED!\n", m_dev->name);

		} else {
			PRINTM(CMD, "BT: %s: HS Enable Fail\n", m_dev->name);
		}
		break;
	case BT_CMD_MODULE_CFG_REQ:
		if ((priv->bt_dev.sendcmdflag == TRUE) &&
			((pEvent->data[1] == MODULE_BRINGUP_REQ)
			 || (pEvent->data[1] == MODULE_SHUTDOWN_REQ))) {
			if (pEvent->data[1] == MODULE_BRINGUP_REQ) {
				PRINTM(CMD, "BT: EVENT %s:%s\n", m_dev->name,
					   (pEvent->data[2] && (pEvent->data[2] !=
						 MODULE_CFG_RESP_ALREADY_UP)) ? "Bring up Fail" : "Bring up success");
				priv->bt_dev.devType = pEvent->data[3];
				PRINTM(CMD, "devType:%s\n",
					   (pEvent->data[3] == DEV_TYPE_AMP) ? "AMP controller" : "BR/EDR controller");
				priv->bt_dev.devFeature = pEvent->data[4];
				PRINTM(CMD, "devFeature:  %s,    %s,    %s,    %s,    %s\n",
						((pEvent->data[4] & DEV_FEATURE_BT) ? "BT Feature" : "No BT Feature"), 
						((pEvent->data[4] & DEV_FEATURE_BTAMP) ? "BTAMP Feature" : "No BTAMP Feature"), 
						((pEvent->data[4] & DEV_FEATURE_BLE) ? "BLE Feature" : "No BLE Feature"),
						((pEvent->data[4] & DEV_FEATURE_FM) ? "FM Feature" : "No FM Feature"),
						((pEvent->data[4] & DEV_FEATURE_NFC) ? "NFC Feature" : "No NFC Feature"));

			}
			if (pEvent->data[1] == MODULE_SHUTDOWN_REQ) {
				PRINTM(CMD, "BT: EVENT %s:%s\n", m_dev->name,
					   (pEvent->data[2]) ? "Shut down Fail" : "Shut down success");

			}
			if (pEvent->data[2]) {
				priv->bt_dev.sendcmdflag = FALSE;
				priv->adapter->cmd_complete = TRUE;
				wake_up_interruptible(&priv->adapter->cmd_wait_q);
			}
		} else {
			PRINTM(CMD, "BT_CMD_MODULE_CFG_REQ resp for APP\n");
			ret = BT_STATUS_FAILURE;
		}
		break;
	case BT_EVENT_POWER_STATE:
		if (pEvent->data[1] == BT_PS_SLEEP)
			priv->adapter->ps_state = PS_SLEEP;
		PRINTM(CMD, "BT: EVENT %s:%s\n", m_dev->name,
			   (priv->adapter->ps_state) ? "PS_SLEEP" : "PS_AWAKE");

		break;
	case BT_CMD_SDIO_PULL_CFG_REQ:
		if (pEvent->data[pEvent->length - 1] == BT_STATUS_SUCCESS)
			PRINTM(CMD, "BT: %s: SDIO pull configuration success\n",
				   m_dev->name);

		else {
			PRINTM(CMD, "BT: %s: SDIO pull configuration fail\n",
				   m_dev->name);

		}
		break;
	default:
		PRINTM(CMD, "BT: Unknown Event=%d %s\n", pEvent->data[0],
			   m_dev->name);
		ret = BT_STATUS_FAILURE;
		break;
	}
exit:
	if (ret == BT_STATUS_SUCCESS)
		kfree_skb(skb);
	LEAVE();
	return ret;
}

/** 
 *  @brief This function sends module cfg cmd to firmware
 *
 *  Command format:
 *  +--------+--------+--------+--------+--------+--------+--------+
 *  |     OCF OGF     | Length |                Data               |
 *  +--------+--------+--------+--------+--------+--------+--------+
 *  |     2-byte      | 1-byte |               4-byte              |
 *  +--------+--------+--------+--------+--------+--------+--------+
 *  
 *  @param priv    A pointer to bt_private structure
 *  @param subcmd  sub command 
 *  @return 	   BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
int bt_send_module_cfg_cmd(bt_private * priv, int subcmd)
{
	struct sk_buff *skb = NULL;
	u8 ret = BT_STATUS_SUCCESS;
	BT_CMD *pCmd;
	ENTER();
	skb = bt_skb_alloc(sizeof(BT_CMD), GFP_ATOMIC);
	if (skb == NULL) {
		PRINTM(WARN, "BT: No free skb\n");
		ret = BT_STATUS_FAILURE;
		goto exit;
	}
	pCmd = (BT_CMD *) skb->data;
	pCmd->ocf_ogf = (OGF << 10) | BT_CMD_MODULE_CFG_REQ;
	pCmd->length = 1;
	pCmd->data[0] = subcmd;
	bt_cb(skb)->pkt_type = MRVL_VENDOR_PKT;
	skb_put(skb, sizeof(BT_CMD));
	skb->dev = (void *) (&(priv->bt_dev.m_dev[BT_SEQ]));
	skb_queue_head(&priv->adapter->tx_queue, skb);
	priv->bt_dev.sendcmdflag = TRUE;
	priv->bt_dev.send_cmd_ocf = BT_CMD_MODULE_CFG_REQ;
	priv->adapter->cmd_complete = FALSE;
	PRINTM(CMD, "Queue module cfg Command(0x%x)\n", pCmd->ocf_ogf);
	wake_up_interruptible(&priv->MainThread.waitQ);
	/* 
	   On some Android platforms certain delay is needed for HCI daemon to
	   remove this module and close itself gracefully. Otherwise it hangs.
	   This 10ms delay is a workaround for such platforms as the root cause
	   has not been found yet.
	*/
	mdelay(10);
	if (!os_wait_interruptible_timeout
		(priv->adapter->cmd_wait_q, priv->adapter->cmd_complete,
		 WAIT_UNTIL_CMD_RESP)) {
		ret = BT_STATUS_FAILURE;
		PRINTM(MSG, "BT: module_cfg_cmd (0x%x): timeout sendcmdflag=%d\n",
			   subcmd, priv->bt_dev.sendcmdflag);
	} else {
		PRINTM(CMD, "BT: module cfg Command done\n");
	}
exit:
	LEAVE();
	return ret;
}

/** 
 *  @brief This function enables power save mode 
 *  
 *  @param priv    A pointer to bt_private structure
 *  @return 	   BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
int bt_enable_ps(bt_private * priv)
{
	struct sk_buff *skb = NULL;
	u8 ret = BT_STATUS_SUCCESS;
	BT_CMD *pCmd;
	ENTER();
	skb = bt_skb_alloc(sizeof(BT_CMD), GFP_ATOMIC);
	if (skb == NULL) {
		PRINTM(WARN, "No free skb\n");
		ret = BT_STATUS_FAILURE;
		goto exit;
	}
	pCmd = (BT_CMD *) skb->data;
	pCmd->ocf_ogf = (OGF << 10) | BT_CMD_AUTO_SLEEP_MODE;
	if (priv->bt_dev.psmode)
		pCmd->data[0] = BT_PS_ENABLE;
	else
		pCmd->data[0] = BT_PS_DISABLE;
	if (priv->bt_dev.idle_timeout) {
		pCmd->length = 3;
		pCmd->data[1] = (u8) (priv->bt_dev.idle_timeout & 0x00ff);
		pCmd->data[2] = (priv->bt_dev.idle_timeout & 0xff00) >> 8;
	} else {
		pCmd->length = 1;
	}
	bt_cb(skb)->pkt_type = MRVL_VENDOR_PKT;
	skb_put(skb, sizeof(BT_CMD));
	skb->dev = (void *) (&(priv->bt_dev.m_dev[BT_SEQ]));
	skb_queue_head(&priv->adapter->tx_queue, skb);
	PRINTM(CMD, "Queue PSMODE Command(0x%x):%d\n", pCmd->ocf_ogf,
		   pCmd->data[0]);
	priv->bt_dev.sendcmdflag = TRUE;
	priv->bt_dev.send_cmd_ocf = BT_CMD_AUTO_SLEEP_MODE;
	priv->adapter->cmd_complete = FALSE;
	wake_up_interruptible(&priv->MainThread.waitQ);
	if (!os_wait_interruptible_timeout
		(priv->adapter->cmd_wait_q, priv->adapter->cmd_complete,
		 WAIT_UNTIL_CMD_RESP)) {
		ret = BT_STATUS_FAILURE;
		PRINTM(MSG, "BT: psmode timeout: %d, %d,%d\n",
			   priv->adapter->cmd_complete, priv->adapter->ps_state,
			   priv->adapter->WakeupTries);
	}
exit:
	LEAVE();
	return ret;
}

/** 
 *  @brief This function sends hscfg command 
 *  
 *  @param priv    A pointer to bt_private structure
 *  @return 	   BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
int bt_send_hscfg_cmd(bt_private * priv)
{
	struct sk_buff *skb = NULL;
	u8 ret = BT_STATUS_SUCCESS;
	BT_CMD *pCmd;
	ENTER();
	skb = bt_skb_alloc(sizeof(BT_CMD), GFP_ATOMIC);
	if (skb == NULL) {
		PRINTM(WARN, "No free skb\n");
		ret = BT_STATUS_FAILURE;
		goto exit;
	}
	pCmd = (BT_CMD *) skb->data;
	pCmd->ocf_ogf = (OGF << 10) | BT_CMD_HOST_SLEEP_CONFIG;
	pCmd->length = 2;
	pCmd->data[0] = (priv->bt_dev.gpio_gap & 0xff00) >> 8;
	pCmd->data[1] = (u8) (priv->bt_dev.gpio_gap & 0x00ff);
	bt_cb(skb)->pkt_type = MRVL_VENDOR_PKT;
	skb_put(skb, sizeof(BT_CMD));
	skb->dev = (void *) (&(priv->bt_dev.m_dev[BT_SEQ]));
	skb_queue_head(&priv->adapter->tx_queue, skb);
	PRINTM(CMD, "Queue HSCFG Command(0x%x),gpio=0x%x,gap=0x%x\n",
		   pCmd->ocf_ogf, pCmd->data[0], pCmd->data[1]);
	priv->bt_dev.sendcmdflag = TRUE;
	priv->bt_dev.send_cmd_ocf = BT_CMD_HOST_SLEEP_CONFIG;
	priv->adapter->cmd_complete = FALSE;
	wake_up_interruptible(&priv->MainThread.waitQ);
	if (!os_wait_interruptible_timeout
		(priv->adapter->cmd_wait_q, priv->adapter->cmd_complete,
		 WAIT_UNTIL_CMD_RESP)) {
		ret = BT_STATUS_FAILURE;
		PRINTM(MSG, "BT: HSCFG timeout: %d, %d,%d\n",
			   priv->adapter->cmd_complete, priv->adapter->ps_state,
			   priv->adapter->WakeupTries);
	}
exit:
	LEAVE();
	return ret;
}

/** 
 *  @brief This function sends sdio pull ctrl command 
 *  
 *  @param priv    A pointer to bt_private structure
 *  @return 	   BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
int bt_send_sdio_pull_ctrl_cmd(bt_private * priv)
{
	struct sk_buff *skb = NULL;
	u8 ret = BT_STATUS_SUCCESS;
	BT_CMD *pCmd;
	ENTER();
	skb = bt_skb_alloc(sizeof(BT_CMD), GFP_ATOMIC);
	if (skb == NULL) {
		PRINTM(WARN, "No free skb\n");
		ret = BT_STATUS_FAILURE;
		goto exit;
	}
	pCmd = (BT_CMD *) skb->data;
	pCmd->ocf_ogf = (OGF << 10) | BT_CMD_SDIO_PULL_CFG_REQ;
		pCmd->length = 4;
	pCmd->data[0] = (priv->bt_dev.sdio_pull_cfg & 0x000000ff);
	pCmd->data[1] = (priv->bt_dev.sdio_pull_cfg & 0x0000ff00) >> 8;
	pCmd->data[2] = (priv->bt_dev.sdio_pull_cfg & 0x00ff0000) >> 16;
	pCmd->data[3] = (priv->bt_dev.sdio_pull_cfg & 0xff000000) >> 24;
	bt_cb(skb)->pkt_type = MRVL_VENDOR_PKT;
	skb_put(skb, sizeof(BT_CMD));
	skb->dev = (void *) (&(priv->bt_dev.m_dev[BT_SEQ]));
	skb_queue_head(&priv->adapter->tx_queue, skb);
	PRINTM(CMD,
		   "Queue SDIO PULL CFG Command(0x%x), PullUp=0x%x%x,PullDown=0x%x%x\n",
		   pCmd->ocf_ogf, pCmd->data[1], pCmd->data[0], pCmd->data[3], pCmd->data[2]);
	priv->bt_dev.sendcmdflag = TRUE;
	priv->bt_dev.send_cmd_ocf = BT_CMD_SDIO_PULL_CFG_REQ;
	priv->adapter->cmd_complete = FALSE;
	wake_up_interruptible(&priv->MainThread.waitQ);
	if (!os_wait_interruptible_timeout
		(priv->adapter->cmd_wait_q, priv->adapter->cmd_complete,
		 WAIT_UNTIL_CMD_RESP)) {
		ret = BT_STATUS_FAILURE;
		PRINTM(MSG, "BT: SDIO PULL CFG timeout: %d, %d,%d\n",
			   priv->adapter->cmd_complete, priv->adapter->ps_state,
			   priv->adapter->WakeupTries);
	}
exit:
	LEAVE();
	return ret;
}

/** 
 *  @brief This function enables host sleep 
 *  
 *  @param priv    A pointer to bt_private structure
 *  @return 	   BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
int bt_enable_hs(bt_private * priv)
{
	struct sk_buff *skb = NULL;
	u8 ret = BT_STATUS_SUCCESS;
	BT_CMD *pCmd;
	ENTER();
	skb = bt_skb_alloc(sizeof(BT_CMD), GFP_ATOMIC);
	if (skb == NULL) {
		PRINTM(WARN, "No free skb\n");
		ret = BT_STATUS_FAILURE;
		goto exit;
	}
	priv->adapter->suspend_fail = FALSE;
	pCmd = (BT_CMD *) skb->data;
	pCmd->ocf_ogf = (OGF << 10) | BT_CMD_HOST_SLEEP_ENABLE;
	pCmd->length = 0;
	bt_cb(skb)->pkt_type = MRVL_VENDOR_PKT;
	skb_put(skb, sizeof(BT_CMD));
	skb->dev = (void *) (&(priv->bt_dev.m_dev[BT_SEQ]));
	skb_queue_head(&priv->adapter->tx_queue, skb);
	priv->bt_dev.sendcmdflag = TRUE;
	priv->bt_dev.send_cmd_ocf = BT_CMD_HOST_SLEEP_ENABLE;
	PRINTM(CMD, "Queue hs enable Command(0x%x)\n", pCmd->ocf_ogf);
	wake_up_interruptible(&priv->MainThread.waitQ);
	os_wait_interruptible_timeout(priv->adapter->cmd_wait_q,
								  priv->adapter->hs_state,
								  WAIT_UNTIL_HS_STATE_CHANGED);
	OS_INT_DISABLE;
	if ((priv->adapter->hs_state == HS_ACTIVATED) ||
		(priv->adapter->is_suspended == TRUE)) {
		OS_INT_RESTORE;
		PRINTM(MSG, "BT: suspend success! skip=%d\n", priv->adapter->hs_skip);
	} else {
		priv->adapter->suspend_fail = TRUE;
		OS_INT_RESTORE;
		priv->adapter->hs_skip++;
		ret = BT_STATUS_FAILURE;
		PRINTM(MSG,
			   "BT: suspend skipped! state=%d skip=%d ps_state= %d WakeupTries=%d\n",
			   priv->adapter->hs_state, priv->adapter->hs_skip,
			   priv->adapter->ps_state, priv->adapter->WakeupTries);
	}
exit:
	LEAVE();
	return ret;
}

/** 
 *  @brief This function sets ble deepsleep mode
 *  
 *  @param priv    A pointer to bt_private structure
 *  @param mode    TRUE/FALSE
 *
 *  @return 	   BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
int bt_set_ble_deepsleep(bt_private * priv, int mode)
{
	struct sk_buff *skb = NULL;
	u8 ret = BT_STATUS_SUCCESS;
	BT_BLE_CMD *pCmd;
	ENTER();
	skb = bt_skb_alloc(sizeof(BT_BLE_CMD), GFP_ATOMIC);
	if (skb == NULL) {
		PRINTM(WARN, "No free skb\n");
		ret = BT_STATUS_FAILURE;
		goto exit;
	}
	pCmd = (BT_BLE_CMD *) skb->data;
	pCmd->ocf_ogf = (OGF << 10) | BT_CMD_BLE_DEEP_SLEEP;
	pCmd->length = 1;
	pCmd->deepsleep = mode;
	bt_cb(skb)->pkt_type = HCI_COMMAND_PKT;
	skb_put(skb, sizeof(BT_BLE_CMD));
	skb->dev = (void *) (&(priv->bt_dev.m_dev[BT_SEQ]));
	skb_queue_head(&priv->adapter->tx_queue, skb);
	priv->bt_dev.sendcmdflag = TRUE;
	priv->bt_dev.send_cmd_ocf = BT_CMD_BLE_DEEP_SLEEP;
	priv->adapter->cmd_complete = FALSE;
	PRINTM(CMD, "BT: Set BLE deepsleep = %d (0x%x)\n", mode, pCmd->ocf_ogf);
	wake_up_interruptible(&priv->MainThread.waitQ);
	if (!os_wait_interruptible_timeout
		(priv->adapter->cmd_wait_q, priv->adapter->cmd_complete,
		 WAIT_UNTIL_CMD_RESP)) {
		ret = BT_STATUS_FAILURE;
		PRINTM(MSG, "BT: Set BLE deepsleep timeout: %d, %d,%d\n",
			   priv->adapter->cmd_complete, priv->adapter->ps_state,
			   priv->adapter->WakeupTries);
	}
exit:
	LEAVE();
	return ret;
}

/** 
 *  @brief This function gets FW version
 *  
 *  @param priv    A pointer to bt_private structure
 *
 *  @return 	   BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
int bt_get_fw_version(bt_private *priv)
{
	struct sk_buff *skb = NULL;
	u8 ret = BT_STATUS_SUCCESS;
	BT_HCI_CMD * pCmd;
	ENTER();
	skb = bt_skb_alloc(sizeof(BT_HCI_CMD), GFP_ATOMIC);
	if(skb == NULL){
		PRINTM(WARN, "No free skb\n");
		ret = BT_STATUS_FAILURE;
		goto exit;
	}
	pCmd = (BT_HCI_CMD *)skb->data;
	pCmd->ocf_ogf = (OGF << 10) | BT_CMD_GET_FW_VERSION;
	pCmd->length = 0x01;
	pCmd->cmd_type = 0x00;
	bt_cb(skb)->pkt_type = HCI_COMMAND_PKT;
	skb_put(skb, 4);
	skb->dev = priv->bt_dev.m_dev[BT_SEQ].dev_pointer;
	skb_queue_head(&priv->adapter->tx_queue, skb);
	priv->bt_dev.sendcmdflag = TRUE;
	priv->bt_dev.send_cmd_ocf = BT_CMD_GET_FW_VERSION;
	priv->adapter->cmd_complete = FALSE;
	wake_up_interruptible(&priv->MainThread.waitQ);
	if (!os_wait_interruptible_timeout(priv->adapter->cmd_wait_q,priv->adapter->cmd_complete,
				WAIT_UNTIL_CMD_RESP)) {
		ret = BT_STATUS_FAILURE;
		PRINTM(MSG, "BT: Get FW version: timeout: %d, %d,%d\n", priv->adapter->cmd_complete,
				priv->adapter->ps_state, priv->adapter->WakeupTries);
	}
exit:
	LEAVE();
	return ret;
}

/** 
 *  @brief This function sets mac address 
 *  
 *  @param priv    A pointer to bt_private structure
 *  @param mac     A pointer to mac address
 *
 *  @return 	   BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
int bt_set_mac_address(bt_private * priv, u8 * mac)
{
	struct sk_buff *skb = NULL;
	u8 ret = BT_STATUS_SUCCESS;
	BT_HCI_CMD *pCmd;
	int i = 0;
	ENTER();
	skb = bt_skb_alloc(sizeof(BT_HCI_CMD), GFP_ATOMIC);
	if (skb == NULL) {
		PRINTM(WARN, "No free skb\n");
		ret = BT_STATUS_FAILURE;
		goto exit;
	}
	pCmd = (BT_HCI_CMD *) skb->data;
	pCmd->ocf_ogf = (OGF << 10) | BT_CMD_CONFIG_MAC_ADDR;
	pCmd->length = 8;
	pCmd->cmd_type = MRVL_VENDOR_PKT;
	pCmd->cmd_len = 6;
	for (i = 0; i < 6; i++)
		pCmd->data[i] = mac[5 - i];
	bt_cb(skb)->pkt_type = HCI_COMMAND_PKT;
	skb_put(skb, sizeof(BT_HCI_CMD));
	skb->dev = (void *) (&(priv->bt_dev.m_dev[BT_SEQ]));
	skb_queue_head(&priv->adapter->tx_queue, skb);
	priv->bt_dev.sendcmdflag = TRUE;
	priv->bt_dev.send_cmd_ocf = BT_CMD_CONFIG_MAC_ADDR;
	priv->adapter->cmd_complete = FALSE;
	PRINTM(CMD, "BT: Set mac address %02x:%02x:%02x:%02x:%02x:%02x (0x%x)\n",
		   mac[0], mac[1], mac[2], mac[3], mac[4], mac[5], pCmd->ocf_ogf);
	wake_up_interruptible(&priv->MainThread.waitQ);
	if (!os_wait_interruptible_timeout
		(priv->adapter->cmd_wait_q, priv->adapter->cmd_complete,
		 WAIT_UNTIL_CMD_RESP)) {
		ret = BT_STATUS_FAILURE;
		PRINTM(MSG, "BT: Set mac address: timeout: %d, %d,%d\n",
			   priv->adapter->cmd_complete, priv->adapter->ps_state,
			   priv->adapter->WakeupTries);
	}
exit:
	LEAVE();
	return ret;
}

/** 
 *  @brief This function writes value to CSU registers
 *  
 *  @param priv    A pointer to bt_private structure
 *  @param type    reg type
 *  @param offset  register address
 *  @param value   register value to write
 *  @return 	   BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
int bt_write_reg(bt_private * priv, u8 type, u32 offset, u16 value)
{
	struct sk_buff *skb = NULL;
	u8 ret = BT_STATUS_SUCCESS;
	BT_CSU_CMD *pCmd;
	ENTER();
	skb = bt_skb_alloc(sizeof(BT_CSU_CMD), GFP_ATOMIC);
	if (skb == NULL) {
		PRINTM(WARN, "No free skb\n");
		ret = BT_STATUS_FAILURE;
		goto exit;
	}
	pCmd = (BT_CSU_CMD *) skb->data;
	pCmd->ocf_ogf = (OGF << 10) | BT_CMD_CSU_WRITE_REG;
	pCmd->length = 7;
	pCmd->type = type;
	pCmd->offset[0] = (offset & 0x000000ff);
	pCmd->offset[1] = (offset & 0x0000ff00) >> 8;
	pCmd->offset[2] = (offset & 0x00ff0000) >> 16;
	pCmd->offset[3] = (offset & 0xff000000) >> 24;
	pCmd->value[0] = (value & 0x00ff);
	pCmd->value[1] = (value & 0xff00) >> 8;
	bt_cb(skb)->pkt_type = HCI_COMMAND_PKT;
	skb_put(skb, sizeof(BT_CSU_CMD));
	skb->dev = (void *) (&(priv->bt_dev.m_dev[BT_SEQ]));
	skb_queue_head(&priv->adapter->tx_queue, skb);
	priv->bt_dev.sendcmdflag = TRUE;
	priv->bt_dev.send_cmd_ocf = BT_CMD_CSU_WRITE_REG;
	priv->adapter->cmd_complete = FALSE;
	PRINTM(CMD, "BT: Set CSU reg type=%d reg=0x%x value=0x%x\n", type, offset,
		   value);
	wake_up_interruptible(&priv->MainThread.waitQ);
	if (!os_wait_interruptible_timeout
		(priv->adapter->cmd_wait_q, priv->adapter->cmd_complete,
		 WAIT_UNTIL_CMD_RESP)) {
		ret = BT_STATUS_FAILURE;
		PRINTM(ERROR, "BT: Set CSU reg timeout: %d, %d, %d\n",
			   priv->adapter->cmd_complete, priv->adapter->ps_state,
			   priv->adapter->WakeupTries);
	}
exit:
	LEAVE();
	return ret;
}

/** 
 *  @brief This function used to restore tx_queue
 *
 *  @param priv    A pointer to bt_private structure
 *  @return        N/A
 */
void bt_restore_tx_queue(bt_private * priv)
{
	struct sk_buff *skb = NULL;
	while (!skb_queue_empty(&priv->adapter->pending_queue)) {
		skb = skb_dequeue(&priv->adapter->pending_queue);
		skb_queue_tail(&priv->adapter->tx_queue, skb);
	}
	wake_up_interruptible(&priv->MainThread.waitQ);
}

/** 
 *  @brief This function used to send command to firmware
 *
 *  Command format:
 *  +--------+--------+--------+--------+--------+--------+--------+
 *  |     OCF OGF     | Length |                Data               |
 *  +--------+--------+--------+--------+--------+--------+--------+
 *  |     2-byte      | 1-byte |               4-byte              |
 *  +--------+--------+--------+--------+--------+--------+--------+
 *  
 *  @param priv    A pointer to bt_private structure
 *  @return        BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
int bt_prepare_command(bt_private * priv)
{
	u8 ret = BT_STATUS_SUCCESS;
	ENTER();
	if (priv->bt_dev.hscfgcmd) {
		priv->bt_dev.hscfgcmd = 0;
		ret = bt_send_hscfg_cmd(priv);
	}
	if (priv->bt_dev.pscmd) {
		priv->bt_dev.pscmd = 0;
		ret = bt_enable_ps(priv);
	}
	if (priv->bt_dev.sdio_pull_ctrl) {
		priv->bt_dev.sdio_pull_ctrl = 0;
		ret = bt_send_sdio_pull_ctrl_cmd(priv);
	}
	if (priv->bt_dev.hscmd) {
		priv->bt_dev.hscmd = 0;
		if (priv->bt_dev.hsmode)
			ret = bt_enable_hs(priv);
		else {
			ret = sbi_wakeup_firmware(priv);
			priv->adapter->hs_state = HS_DEACTIVATED;
		}
	}
	LEAVE();
	return ret;
}

/** @brief This function processes a single packet 
 *  
 *  @param priv    A pointer to bt_private structure
 *  @param skb     A pointer to skb which includes TX packet
 *  @return 	   BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
static int SendSinglePacket(bt_private * priv, struct sk_buff *skb)
{
	int ret;
	ENTER();
	if (!skb || !skb->data) {
		LEAVE();
		return BT_STATUS_FAILURE;
	}
	if (!skb->len || ((skb->len + BT_HEADER_LEN) > BT_UPLD_SIZE)) {
		PRINTM(ERROR, "Tx Error: Bad skb length %d : %d\n", skb->len,
			   BT_UPLD_SIZE);
		LEAVE();
		return BT_STATUS_FAILURE;
	}
	if (skb_headroom(skb) < BT_HEADER_LEN) {
		struct sk_buff *tmp = skb;
		skb = skb_realloc_headroom(skb, BT_HEADER_LEN);
		if (!skb) {
			PRINTM(ERROR, "TX error: realloc_headroom failed %d\n",
					BT_HEADER_LEN);
			skb = tmp;
			LEAVE();
			return BT_STATUS_FAILURE;
		}
		kfree_skb(tmp);
	}
	/* This is SDIO specific header length: byte[3][2][1], * type: byte[0]
	   (HCI_COMMAND = 1, ACL_DATA = 2, SCO_DATA = 3, 0xFE = Vendor) */
	skb_push(skb, BT_HEADER_LEN);
	skb->data[0] = (skb->len & 0x0000ff);
	skb->data[1] = (skb->len & 0x00ff00) >> 8;
	skb->data[2] = (skb->len & 0xff0000) >> 16;
	skb->data[3] = bt_cb(skb)->pkt_type;
	if (bt_cb(skb)->pkt_type == MRVL_VENDOR_PKT)
		PRINTM(CMD, "DNLD_CMD: ocf_ogf=0x%x len=%d\n",
			   *((u16 *) & skb->data[4]), skb->len);
	ret = sbi_host_to_card(priv, skb->data, skb->len);
	LEAVE();
	return ret;
}

/**
 *  @brief This function initializes the adapter structure
 *  and set default value to the member of adapter.
 *  
 *  @param priv    A pointer to bt_private structure
 *  @return 	   N/A
 */
static void bt_init_adapter(bt_private * priv)
{
	ENTER();
	skb_queue_head_init(&priv->adapter->tx_queue);
	skb_queue_head_init(&priv->adapter->pending_queue);
	priv->adapter->tx_lock = FALSE;
	priv->adapter->ps_state = PS_AWAKE;
	priv->adapter->suspend_fail = FALSE;
	priv->adapter->is_suspended = FALSE;
	priv->adapter->hs_skip = 0;
	init_waitqueue_head(&priv->adapter->cmd_wait_q);
	LEAVE();
}

/** 
 *  @brief This function initializes firmware
 *  
 *  @param priv    A pointer to bt_private structure
 *  @return 	   BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
static int bt_init_fw(bt_private * priv)
{
	int ret = BT_STATUS_SUCCESS;
	ENTER();
	if (fw == 0) {
		sbi_enable_host_int(priv);
		goto done;
	}
	sbi_disable_host_int(priv);
	if (sbi_download_fw(priv)) {
		PRINTM(ERROR, " FW failed to be download!\n");
		ret = BT_STATUS_FAILURE;
		goto done;
	}
done:
	LEAVE();
	return ret;
}

/** 
 *  @brief This function frees the structure of adapter
 *    
 *  @param priv    A pointer to bt_private structure
 *  @return 	   N/A
 */
void bt_free_adapter(bt_private * priv)
{
	bt_adapter *Adapter = priv->adapter;
	ENTER();
	skb_queue_purge(&priv->adapter->tx_queue);
	/* Free the adapter object itself */
	kfree(Adapter);
	priv->adapter = NULL;

	LEAVE();
}

/** 
 *  @brief This function handles the BT ioctl 
 *  
 *  @param hdev     A pointer to hci_dev structure
 *  @cmd   	   ioctl cmd 
 *  @arg   	   argument
 *  @return 	   -ENOIOCTLCMD
 */
static int bt_ioctl(struct hci_dev *hdev, unsigned int cmd, unsigned long arg)
{
	ENTER();
	LEAVE();
	return -ENOIOCTLCMD;
}

/** 
 *  @brief This function handles the wrapper_dev ioctl 
 *  
 *  @param hev     A pointer to wrapper_dev structure
 *  @cmd   	   ioctl cmd 
 *  @arg   	   argument
 *  @return 	   -ENOIOCTLCMD
 */
static int mdev_ioctl(struct m_dev *m_dev, unsigned int cmd,
					  unsigned long arg)
{
	ENTER();
	LEAVE();
	return -ENOIOCTLCMD;
}

/** 
 *  @brief This function handles BT destruct 
 *  
 *  @param hdev    A pointer to hci_dev structure
 *  
 *  @return 	   N/A
 */
 #if LINUX_VERSION_CODE < KERNEL_VERSION(3,4,0) 
static void bt_destruct(struct hci_dev *hdev)
{
	ENTER();
	LEAVE();
	return;
}
#endif

/** 
 *  @brief This function handles wrapper device destruct 
 *  
 *  @param hdev    A pointer to m_dev structure
 *  
 *  @return 	   N/A
 */
static void mdev_destruct(struct m_dev *m_dev)
{
	ENTER();
	LEAVE();
	return;
}

/** 
 *  @brief This function handles the BT transmit
 *  
 *  @param skb     A pointer to sk_buff structure
 *  
 *  @return 	   BT_STATUS_SUCCESS or other error no.   
 */
static int bt_send_frame(struct sk_buff *skb)
{
	struct hci_dev *hdev = (struct hci_dev *) skb->dev;
	bt_private *priv = NULL;

	ENTER();
	PRINTM(DATA, "bt_send_frame %s: Type=%d, len=%d\n", hdev->name,
		   bt_cb(skb)->pkt_type, skb->len);
	DBG_HEXDUMP(CMD_D, "bt_send_frame", skb->data, skb->len);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,4,0)	
	if (!hdev || !hdev->core_data) {
#else
	if (!hdev || !hdev->driver_data) {
#endif
		PRINTM(ERROR, "Frame for unknown HCI device (hdev=NULL)\n");
		LEAVE();
		return -ENODEV;
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,4,0)
	priv = (bt_private *) hdev->core_data;
#else
	priv = (bt_private *) hdev->driver_data;
#endif
	if (!test_bit(HCI_RUNNING, &hdev->flags)) {
		PRINTM(ERROR, "Fail test HCI_RUNING, flag=0x%lx\n", hdev->flags);
		LEAVE();
		return -EBUSY;
	}
	switch (bt_cb(skb)->pkt_type) {
	case HCI_COMMAND_PKT:
		hdev->stat.cmd_tx++;
		break;
	case HCI_ACLDATA_PKT:
		hdev->stat.acl_tx++;
		break;
	case HCI_SCODATA_PKT:
		hdev->stat.sco_tx++;
		break;
	}
	if (priv->adapter->tx_lock == TRUE)
		skb_queue_tail(&priv->adapter->pending_queue, skb);
	else
		skb_queue_tail(&priv->adapter->tx_queue, skb);
	wake_up_interruptible(&priv->MainThread.waitQ);
	LEAVE();
	return BT_STATUS_SUCCESS;
}

/** 
 *  @brief This function handles the wrapper device transmit
 *  
 *  @param skb     A pointer to sk_buff structure
 *  
 *  @return 	   BT_STATUS_SUCCESS or other error no.   
 */
static int mdev_send_frame(struct m_dev *m_dev, struct sk_buff *skb)
{
	bt_private *priv = NULL;

	ENTER();
	if (!m_dev || !m_dev->driver_data) {
		PRINTM(ERROR, "Frame for unknown HCI device (m_dev=NULL)\n");
		LEAVE();
		return -ENODEV;
	}
	priv = (bt_private *) m_dev->driver_data;
	if (!test_bit(HCI_RUNNING, &m_dev->flags)) {
		PRINTM(ERROR, "Fail test HCI_RUNING, flag=0x%lx\n", m_dev->flags);
		LEAVE();
		return -EBUSY;
	}
	switch (bt_cb(skb)->pkt_type) {
	case HCI_COMMAND_PKT:
		m_dev->stat.cmd_tx++;
		break;
	case HCI_ACLDATA_PKT:
		m_dev->stat.acl_tx++;
		break;
	case HCI_SCODATA_PKT:
		m_dev->stat.sco_tx++;
		break;
	}

	if (priv->adapter->tx_lock == TRUE)
		skb_queue_tail(&priv->adapter->pending_queue, skb);
	else
		skb_queue_tail(&priv->adapter->tx_queue, skb);
	wake_up_interruptible(&priv->MainThread.waitQ);

	LEAVE();
	return BT_STATUS_SUCCESS;
}

/** 
 *  @brief This function flushes the transmit queue
 *  
 *  @param hdev     A pointer to hci_dev structure
 *  
 *  @return 	   BT_STATUS_SUCCESS   
 */
static int bt_flush(struct hci_dev *hdev)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,4,0)
	bt_private *priv = (bt_private *) hdev->core_data;
#else
	bt_private *priv = (bt_private *) hdev->driver_data;
#endif
	ENTER();
		skb_queue_purge(&priv->adapter->tx_queue);
	skb_queue_purge(&priv->adapter->pending_queue);
	LEAVE();
	return BT_STATUS_SUCCESS;
}

/** 
 *  @brief This function flushes the transmit queue
 *  
 *  @param m_dev     A pointer to m_dev structure
 *  
 *  @return 	   BT_STATUS_SUCCESS   
 */
static int mdev_flush(struct m_dev *m_dev)
{
	bt_private *priv = (bt_private *) m_dev->driver_data;
	ENTER();
		skb_queue_purge(&priv->adapter->tx_queue);
	skb_queue_purge(&priv->adapter->pending_queue);
	LEAVE();
	return BT_STATUS_SUCCESS;
}

/** 
 *  @brief This function closes the bluetooth device
 *  
 *  @param hdev    A pointer to hci_dev structure
 *  
 *  @return 	   BT_STATUS_SUCCESS   
 */
static int bt_close(struct hci_dev *hdev)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,4,0)
	bt_private *priv = (bt_private *) hdev->core_data;
#else
	bt_private *priv = (bt_private *) hdev->driver_data;
#endif
	ENTER();
	if (!test_and_clear_bit(HCI_RUNNING, &hdev->flags)) {
		LEAVE();
		return BT_STATUS_SUCCESS;
	}
		skb_queue_purge(&priv->adapter->tx_queue);
	module_put(THIS_MODULE);
	LEAVE();
	return BT_STATUS_SUCCESS;
}

/** 
 *  @brief This function closes the wrapper device
 *  
 *  @param m_dev   A pointer to m_dev structure
 *  
 *  @return 	   BT_STATUS_SUCCESS   
 */
static int mdev_close(struct m_dev *m_dev)
{
	ENTER();
	mdev_req_lock(m_dev);
	if (!test_and_clear_bit(HCI_UP, &m_dev->flags)) {
		mdev_req_unlock(m_dev);
		LEAVE();
		return 0;
	}

	if (m_dev->flush)
		m_dev->flush(m_dev);
	/* wait up pending read and unregister char dev */
	wake_up_interruptible(&m_dev->req_wait_q);
	/* Drop queues */
	skb_queue_purge(&m_dev->rx_q);

	if (!test_and_clear_bit(HCI_RUNNING, &m_dev->flags)) {
		mdev_req_unlock(m_dev);
		LEAVE();
		return 0;
	}
	module_put(THIS_MODULE);
	m_dev->flags = 0;
	mdev_req_unlock(m_dev);
	LEAVE();
	return BT_STATUS_SUCCESS;
}

/** 
 *  @brief This function opens the bluetooth device
 *  
 *  @param hdev    A pointer to hci_dev structure
 *  
 *  @return 	   BT_STATUS_SUCCESS  or other
 */
static int bt_open(struct hci_dev *hdev)
{
	ENTER();
	if (try_module_get(THIS_MODULE) == 0)
		return BT_STATUS_FAILURE;
	set_bit(HCI_RUNNING, &hdev->flags);
	LEAVE();
	return BT_STATUS_SUCCESS;
}

/** 
 *  @brief This function opens the wrapper device
 *  
 *  @param m_dev   A pointer to m_dev structure
 *  
 *  @return 	   BT_STATUS_SUCCESS  or other
 */
static int mdev_open(struct m_dev *m_dev)
{
	ENTER();

	if (try_module_get(THIS_MODULE) == 0) {
		return BT_STATUS_FAILURE;
	}
	set_bit(HCI_RUNNING, &m_dev->flags);

	LEAVE();
	return BT_STATUS_SUCCESS;
}

/** 
 *  @brief This function queries the wrapper device
 *  
 *  @param m_dev   A pointer to m_dev structure
 *  @param arg     arguement
 *  
 *  @return 	   BT_STATUS_SUCCESS  or other
 */
void mdev_query(struct m_dev *m_dev, unsigned long arg)
{
	struct mbt_dev *mbt_dev = (struct mbt_dev *) m_dev->dev_pointer;

	ENTER();
	if (copy_to_user((void *) arg, &mbt_dev->type, sizeof(mbt_dev->type)))
		PRINTM(ERROR, "IOCTL_QUERY_TYPE: Fail copy to user\n");

	LEAVE();
}

/** 
 *  @brief This function initializes the wrapper device
 *  
 *  @param m_dev   A pointer to m_dev structure
 *  
 *  @return 	   BT_STATUS_SUCCESS  or other
 */
void init_m_dev(struct m_dev *m_dev)
{
	m_dev->dev_pointer = NULL;
	m_dev->driver_data = NULL;
	m_dev->dev_type = 0;
	m_dev->spec_type = 0;
	skb_queue_head_init(&m_dev->rx_q);
	init_waitqueue_head(&m_dev->req_wait_q);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,37)
	init_MUTEX(&m_dev->req_lock);
#else
	sema_init(&m_dev->req_lock, 1);
#endif
	memset(&m_dev->stat, 0, sizeof(struct hci_dev_stats));
	m_dev->open = mdev_open;
	m_dev->close = mdev_close;
	m_dev->flush = mdev_flush;
	m_dev->send = mdev_send_frame;
	m_dev->destruct = mdev_destruct;
	m_dev->ioctl = mdev_ioctl;
	m_dev->query = mdev_query;
	m_dev->owner = THIS_MODULE;

}

/** 
 *  @brief This function handles the major job in bluetooth driver.
 *  it handles the event generated by firmware, rx data received
 *  from firmware and tx data sent from kernel.
 *  
 *  @param data    A pointer to bt_thread structure
 *  @return        BT_STATUS_SUCCESS
 */
static int bt_service_main_thread(void *data)
{
	bt_thread *thread = data;
	bt_private *priv = thread->priv;
	bt_adapter *Adapter = priv->adapter;
	wait_queue_t wait;
	struct sk_buff *skb;
	ENTER();
	bt_activate_thread(thread);
	init_waitqueue_entry(&wait, current);
	current->flags |= PF_NOFREEZE;

	for (;;) {
		add_wait_queue(&thread->waitQ, &wait);
		OS_SET_THREAD_STATE(TASK_INTERRUPTIBLE);
		if (priv->adapter->WakeupTries ||
			((!priv->adapter->IntCounter) &&
			 (!priv->bt_dev.tx_dnld_rdy ||
			  skb_queue_empty(&priv->adapter->tx_queue)))) {
			PRINTM(INFO, "Main: Thread sleeping...\n");
			schedule();
		}
		OS_SET_THREAD_STATE(TASK_RUNNING);
		remove_wait_queue(&thread->waitQ, &wait);
		if (kthread_should_stop() || Adapter->SurpriseRemoved) {
			PRINTM(INFO, "main-thread: break from main thread: "
				   "SurpriseRemoved=0x%x\n", Adapter->SurpriseRemoved);
			break;
		}

		PRINTM(INFO, "Main: Thread waking up...\n");
		if (priv->adapter->IntCounter) {
			OS_INT_DISABLE;
			Adapter->IntCounter = 0;
			OS_INT_RESTORE;
			sbi_get_int_status(priv);
		} else if ((priv->adapter->ps_state == PS_SLEEP) &&
				   !skb_queue_empty(&priv->adapter->tx_queue)) {
			priv->adapter->WakeupTries++;
			sbi_wakeup_firmware(priv);
			continue;
		}
		if (priv->adapter->ps_state == PS_SLEEP)
			continue;
		if (priv->bt_dev.tx_dnld_rdy == TRUE) {
			if (!skb_queue_empty(&priv->adapter->tx_queue)) {
				skb = skb_dequeue(&priv->adapter->tx_queue);
				if (skb) {
					if (SendSinglePacket(priv, skb)) {
						if (((priv->bt_dev.m_dev[FM_SEQ].dev_pointer) && (!memcmp(skb->dev, priv->bt_dev.m_dev[FM_SEQ].name, FM_CHAR_NAME_LEN))) ||
							((priv->bt_dev.m_dev[NFC_SEQ].dev_pointer) && (!memcmp(skb->dev, priv->bt_dev.m_dev[NFC_SEQ].name, NFC_CHAR_NAME_LEN))))
						{
							((struct m_dev *)skb->dev)->stat.err_tx ++;
						}
						else ((struct hci_dev *)skb->dev)->stat.err_tx ++;
					}
					else {
						if (((priv->bt_dev.m_dev[FM_SEQ].dev_pointer) && (!memcmp(skb->dev, priv->bt_dev.m_dev[FM_SEQ].name, FM_CHAR_NAME_LEN))) ||
							((priv->bt_dev.m_dev[NFC_SEQ].dev_pointer) && (!memcmp(skb->dev, priv->bt_dev.m_dev[NFC_SEQ].name, NFC_CHAR_NAME_LEN))))
						{
							((struct m_dev *)skb->dev)->stat.byte_tx += skb->len;
						}
						else ((struct hci_dev *)skb->dev)->stat.byte_tx += skb->len;
					}
					kfree_skb(skb);
				}
			}
		}
	}
	bt_deactivate_thread(thread);
	LEAVE();
	return BT_STATUS_SUCCESS;
}

/** 
 *  @brief This function handles the interrupt. it will change PS
 *  state if applicable. it will wake up main_thread to handle
 *  the interrupt event as well.
 *  
 *  @param m_dev   A pointer to m_dev structure
 *  @return        N/A
 */
void bt_interrupt(struct m_dev *m_dev)
{
	bt_private *priv = (bt_private *) m_dev->driver_data;
	ENTER();
	if (!priv || !priv->adapter) {
		LEAVE();
		return;
	}
	PRINTM(INTR, "*\n");
	priv->adapter->ps_state = PS_AWAKE;
	if (priv->adapter->hs_state == HS_ACTIVATED) {
		PRINTM(CMD, "BT: %s: HS DEACTIVATED in ISR!\n", m_dev->name);
		priv->adapter->hs_state = HS_DEACTIVATED;
	}
	priv->adapter->WakeupTries = 0;
	priv->adapter->IntCounter++;
	wake_up_interruptible(&priv->MainThread.waitQ);
	LEAVE();
}

/** 
 *  @brief Module configuration and register device
 *  
 *  @param priv      A Pointer to bt_private structure
 *  @return 	     BT_STATUS_SUCCESS or BT_STATUS_FAILURE	
 */
int sd_register_conf_dpc(bt_private * priv)
{
	int ret = BT_STATUS_SUCCESS;
	struct mbt_dev *mbt_dev = NULL;
	struct fm_dev *fm_dev = NULL;
	struct nfc_dev *nfc_dev = NULL;
	struct hci_dev *hdev = NULL;
	struct char_dev *char_dev = NULL;
	char dev_file[16];

	ENTER();

	priv->bt_dev.tx_dnld_rdy = TRUE;

	if (drv_mode & DRV_MODE_BT) {
		hdev = hci_alloc_dev();
		if (!hdev) {
			PRINTM(FATAL, "Can not allocate HCI device\n");
			ret = -ENOMEM;
			goto err_kmalloc;
		}
		hdev->open = bt_open;
		hdev->close = bt_close;
		hdev->flush = bt_flush;
		hdev->send = bt_send_frame;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,4,0)
		hdev->destruct = bt_destruct;
#endif
		hdev->ioctl = bt_ioctl;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,4,0)
		hdev->core_data = priv;
#else
		hdev->driver_data = priv;
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,4,0)
		hdev->owner = THIS_MODULE;
#endif
		priv->bt_dev.m_dev[BT_SEQ].dev_type = BT_TYPE;
		priv->bt_dev.m_dev[BT_SEQ].spec_type = BLUEZ_SPEC;
		priv->bt_dev.m_dev[BT_SEQ].dev_pointer = (void *) hdev;
		priv->bt_dev.m_dev[BT_SEQ].driver_data = priv;
	}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,34)
	if (hdev)
		hdev->bus = HCI_SDIO;
#else
	if (hdev)
		hdev->type = HCI_SDIO;
#endif /* >= 2.6.34 */

	ret = bt_send_module_cfg_cmd(priv, MODULE_BRINGUP_REQ);
	if (ret < 0) {
		PRINTM(FATAL, "Module cfg command send failed!\n");
		goto done;
	}
	ret = bt_set_ble_deepsleep(priv, TRUE);
	if (ret < 0) {
		PRINTM(FATAL, "Enable BLE deepsleep failed!\n");
		goto done;
	}
	if (psmode) {
		priv->bt_dev.psmode = TRUE;
		priv->bt_dev.idle_timeout = DEFAULT_IDLE_TIME;
		ret = bt_enable_ps(priv);
		if (ret < 0) {
			PRINTM(FATAL, "Enable PS mode failed!\n");
			goto done;
		}
	}
#ifdef SDIO_SUSPEND_RESUME
	priv->bt_dev.gpio_gap = 0xffff;
	ret = bt_send_hscfg_cmd(priv);
	if (ret < 0) {
		PRINTM(FATAL, "Send HSCFG failed!\n");
		goto done;
	}
#endif
	priv->bt_dev.sdio_pull_cfg = 0xffffffff;
	priv->bt_dev.sdio_pull_ctrl = 0;
	wake_up_interruptible(&priv->MainThread.waitQ);
	/* block all the packet from bluez */
	if (init_cfg)
		priv->adapter->tx_lock = TRUE;

	if (hdev) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,34)
		hdev->dev_type = priv->bt_dev.devType;
#endif
		ret = hci_register_dev(hdev);
	}
	if (ret < 0) {
		PRINTM(FATAL, "Can not register HCI device\n");
		ret = BT_STATUS_FAILURE;
		goto done;
	}
	snprintf(priv->bt_dev.m_dev[BT_SEQ].name,
			 sizeof(priv->bt_dev.m_dev[BT_SEQ].name), hdev->name);
	bt_proc_init(priv, &(priv->bt_dev.m_dev[BT_SEQ]), BT_SEQ);


	if ((drv_mode & DRV_MODE_FM) && (!(priv->bt_dev.devType == DEV_TYPE_AMP)) && (priv->bt_dev.devFeature & DEV_FEATURE_FM)) {
		/** alloc fm_dev */
		fm_dev = alloc_fm_dev();
		if (!fm_dev) {
			PRINTM(FATAL, "Can not allocate fm dev\n");
			ret = -ENOMEM;
			goto err_kmalloc;
		}

		/** init m_dev */
		init_m_dev(&(priv->bt_dev.m_dev[FM_SEQ]));
		priv->bt_dev.m_dev[FM_SEQ].dev_type = FM_TYPE;
		priv->bt_dev.m_dev[FM_SEQ].spec_type = GENERIC_SPEC;
		priv->bt_dev.m_dev[FM_SEQ].dev_pointer = (void *) fm_dev;
		priv->bt_dev.m_dev[FM_SEQ].driver_data = priv;

		/** create char device for FM */
		char_dev = kzalloc(sizeof(struct char_dev), GFP_KERNEL);
		if (!char_dev) {
			class_destroy(chardev_class);
			ret = -ENOMEM;
			goto err_kmalloc;
		}
		char_dev->minor = FMCHAR_MINOR_BASE + fmchar_minor;
		char_dev->dev_type = FM_TYPE;
		snprintf(fm_dev->name, sizeof(fm_dev->name), "mfmchar%d",
				 fmchar_minor);
		snprintf(dev_file, sizeof(dev_file), "/dev/mfmchar%d", fmchar_minor);
		fmchar_minor++;

		/** register char dev */
		register_char_dev(char_dev, chardev_class, MODULE_NAME, fm_dev->name);

		/** chmod for FM char device */
		mbtchar_chmod(dev_file, 0666);

		/** register m_dev to FM char device */
		priv->bt_dev.m_dev[FM_SEQ].index = char_dev->minor;
		char_dev->m_dev = &(priv->bt_dev.m_dev[FM_SEQ]);

		/** create proc device */
		snprintf(priv->bt_dev.m_dev[FM_SEQ].name,
				 sizeof(priv->bt_dev.m_dev[FM_SEQ].name), fm_dev->name);
		bt_proc_init(priv, &(priv->bt_dev.m_dev[FM_SEQ]), FM_SEQ);
	}

	if ((drv_mode & DRV_MODE_NFC) && (!(priv->bt_dev.devType == DEV_TYPE_AMP)) && (priv->bt_dev.devFeature & DEV_FEATURE_NFC)) {
		/** alloc nfc_dev */
		nfc_dev = alloc_nfc_dev();
		if (!nfc_dev) {
			PRINTM(FATAL, "Can not allocate nfc dev\n");
			ret = -ENOMEM;
			goto err_kmalloc;
		}

		/** init m_dev */
		init_m_dev(&(priv->bt_dev.m_dev[NFC_SEQ]));
		priv->bt_dev.m_dev[NFC_SEQ].dev_type = NFC_TYPE;
		priv->bt_dev.m_dev[NFC_SEQ].spec_type = GENERIC_SPEC;
		priv->bt_dev.m_dev[NFC_SEQ].dev_pointer = (void *) nfc_dev;
		priv->bt_dev.m_dev[NFC_SEQ].driver_data = priv;

		/** create char device for NFC */
		char_dev = kzalloc(sizeof(struct char_dev), GFP_KERNEL);
		if (!char_dev) {
			class_destroy(chardev_class);
			ret = -ENOMEM;
			goto err_kmalloc;
		}
		char_dev->minor = NFCCHAR_MINOR_BASE + nfcchar_minor;
		char_dev->dev_type = NFC_TYPE;
		snprintf(nfc_dev->name, sizeof(nfc_dev->name), "mnfcchar%d",
				 nfcchar_minor);
		snprintf(dev_file, sizeof(dev_file), "/dev/mnfcchar%d",
				 nfcchar_minor);
		nfcchar_minor++;

		/** register char dev */
		register_char_dev(char_dev, chardev_class, MODULE_NAME,
						  nfc_dev->name);

		/** chmod for NFC char device */
		mbtchar_chmod(dev_file, 0666);

		/** register m_dev to NFC char device */
		priv->bt_dev.m_dev[NFC_SEQ].index = char_dev->minor;
		char_dev->m_dev = &(priv->bt_dev.m_dev[NFC_SEQ]);

		/** create proc device */
		snprintf(priv->bt_dev.m_dev[NFC_SEQ].name,
				 sizeof(priv->bt_dev.m_dev[NFC_SEQ].name), nfc_dev->name);
		bt_proc_init(priv, &(priv->bt_dev.m_dev[NFC_SEQ]), NFC_SEQ);
	}

	if (init_cfg) {
		if (BT_STATUS_SUCCESS != bt_init_config(priv, init_cfg)) {
			PRINTM(FATAL, "BT: Set user init data and param failed\n");
			if (hdev) {
				hci_unregister_dev(hdev);
			}
			ret = BT_STATUS_FAILURE;
			goto done;
		} else {
			priv->adapter->tx_lock = FALSE;
			bt_restore_tx_queue(priv);
		}
	}

    /* Get FW version */
        bt_get_fw_version(priv);
    snprintf(priv->adapter->drv_ver, MAX_VER_STR_LEN, driver_version, fw_version);

done:
	LEAVE();
	return ret;
err_kmalloc:
	if (hdev)
		kfree(hdev);
	if (mbt_dev)
		kfree(mbt_dev);
	if (fm_dev)
		kfree(fm_dev);
	if (nfc_dev)
		kfree(nfc_dev);
	LEAVE();
	return ret;
}

/**
 *  @brief This function adds the card. it will probe the
 *  card, allocate the bt_priv and initialize the device. 
 *  
 *  @param card    A pointer to card
 *  @return        A pointer to bt_private structure
 */

bt_private *bt_add_card(void *card)
{
	bt_private *priv = NULL;

	ENTER();

	priv = kzalloc(sizeof(bt_private), GFP_KERNEL);
	if (!priv) {
		PRINTM(FATAL, "Can not allocate priv\n");
		LEAVE();
		return NULL;
	}
	/* allocate buffer for bt_adapter */
	if (!(priv->adapter = kzalloc(sizeof(bt_adapter), GFP_KERNEL))) {
		PRINTM(FATAL, "Allocate buffer for bt_adapter failed!\n");
		goto err_kmalloc;
	}

	bt_init_adapter(priv);

	PRINTM(INFO, "Starting kthread...\n");
	priv->MainThread.priv = priv;
	spin_lock_init(&priv->driver_lock);

	bt_create_thread(bt_service_main_thread, &priv->MainThread,
					 "bt_main_service");

	/* wait for mainthread to up */
	while (!priv->MainThread.pid) {
		os_sched_timeout(1);
	}

	priv->bt_dev.card = card;

	((struct sdio_mmc_card *) card)->priv = priv;
	priv->adapter->sd_ireg = 0;
	/* 
	 * Register the device. Fillup the private data structure with
	 * relevant information from the card and request for the required
	 * IRQ. 
	 */
	if (sbi_register_dev(priv) < 0) {
		PRINTM(FATAL, "Failed to register bt device!\n");
		goto err_registerdev;
	}
	if (bt_init_fw(priv)) {
		PRINTM(FATAL, "BT Firmware Init Failed\n");
		goto err_init_fw;
	}
	LEAVE();
	return priv;

err_init_fw:
	PRINTM(INFO, "unregister device\n");
	sbi_unregister_dev(priv);
err_registerdev:
	((struct sdio_mmc_card *) card)->priv = NULL;
	/* Stop the thread servicing the interrupts */
	priv->adapter->SurpriseRemoved = TRUE;
	wake_up_interruptible(&priv->MainThread.waitQ);
	while (priv->MainThread.pid) {
		os_sched_timeout(1);
	}
err_kmalloc:
	if (priv->adapter)
		bt_free_adapter(priv);
	kfree(priv);
	LEAVE();
	return NULL;
}

/** 
 *  @brief This function removes the card.
 *  
 *  @param card    A pointer to card
 *  @return        BT_STATUS_SUCCESS
 */
int bt_remove_card(void *card)
{
	struct m_dev *m_dev = NULL;
	struct hci_dev *hdev = NULL;
	bt_private *priv = (bt_private *) card;

	ENTER();
	if (!priv) {
		LEAVE();
		return BT_STATUS_SUCCESS;
	}
	if (!priv->adapter->SurpriseRemoved) {
		bt_send_module_cfg_cmd(priv, MODULE_SHUTDOWN_REQ);
		/* Disable interrupts on the card */
		sbi_disable_host_int(priv);
		priv->adapter->SurpriseRemoved = TRUE;
	}

	wake_up_interruptible(&priv->adapter->cmd_wait_q);
	priv->adapter->SurpriseRemoved = TRUE;
	wake_up_interruptible(&priv->MainThread.waitQ);
	while (priv->MainThread.pid) {
		os_sched_timeout(1);
		wake_up_interruptible(&priv->MainThread.waitQ);
	}

	bt_proc_remove(priv);

	PRINTM(INFO, "SD: unregister dev\n");
	sbi_unregister_dev(priv);

	if (priv->bt_dev.m_dev[BT_SEQ].dev_pointer) {
		m_dev = &(priv->bt_dev.m_dev[BT_SEQ]);
		if (m_dev->spec_type == BLUEZ_SPEC) {
			hdev = (struct hci_dev *) m_dev->dev_pointer;
			hci_unregister_dev(hdev);
			hci_free_dev(hdev);
		}
	}
	if (priv->bt_dev.m_dev[FM_SEQ].dev_pointer) {
		m_dev = &(priv->bt_dev.m_dev[FM_SEQ]);
		if ((drv_mode & DRV_MODE_FM) && (fmchar_minor > 0))
			fmchar_minor--;
			/** unregister m_dev to char_dev */
		if (chardev_class)
			chardev_cleanup_one(m_dev, chardev_class);
		free_m_dev(m_dev);
	}
	if (priv->bt_dev.m_dev[NFC_SEQ].dev_pointer) {
		m_dev = &(priv->bt_dev.m_dev[NFC_SEQ]);
		if ((drv_mode & DRV_MODE_NFC) && (nfcchar_minor > 0))
			nfcchar_minor--;
		/** unregister m_dev to char_dev */
		if (chardev_class)
			chardev_cleanup_one(m_dev, chardev_class);
		free_m_dev(m_dev);
	}
	PRINTM(INFO, "SD: Free Adapter\n");
	bt_free_adapter(priv);
	kfree(priv);

	LEAVE();
	return BT_STATUS_SUCCESS;
}


/** 
 *  @brief This function initializes module.
 *  
 *  @return    BT_STATUS_SUCCESS or BT_STATUS_FAILURE
 */
static int bt_init_module(void)
{
	int ret = BT_STATUS_SUCCESS;
	ENTER();
	bt_root_proc_init();
	if (ret)
		goto done;

		/** create char device class */
	chardev_class = class_create(THIS_MODULE, MODULE_NAME);
	if (IS_ERR(chardev_class)) {
		PRINTM(ERROR, "Unable to allocate class\n");
		bt_root_proc_remove();
		ret = PTR_ERR(chardev_class);
		goto done;
	}

	if (sbi_register() == NULL) {
		ret = BT_STATUS_FAILURE;
		goto done;
	}
done:
	LEAVE();
	return ret;
}

/** 
 *  @brief This function cleans module
 *  
 *  @return        N/A
 */
static void bt_exit_module(void)
{
	ENTER();
	printk(KERN_ERR "+++++ enter func bt_exit_module() +++++ \n");
	sbi_unregister();

	bt_root_proc_remove();
	class_destroy(chardev_class);
	LEAVE();
}

module_init(bt_init_module);
module_exit(bt_exit_module);

MODULE_AUTHOR("Marvell International Ltd.");
MODULE_DESCRIPTION("Marvell Bluetooth Driver Ver. " VERSION);
MODULE_VERSION(VERSION);
MODULE_LICENSE("GPL");
module_param(fw, int, 1);
MODULE_PARM_DESC(fw,
				 "0: Skip firmware download; otherwise: Download firmware");
module_param(psmode, int, 1);
MODULE_PARM_DESC(psmode, "1: Enable powermode; 0: Disable powermode");
#ifdef	DEBUG_LEVEL1
module_param(drvdbg, uint, 0);
MODULE_PARM_DESC(drvdbg, "BIT3:DBG_DATA BIT4:DBG_CMD 0xFF:DBG_ALL");
#endif
#ifdef SDIO_SUSPEND_RESUME
module_param(pm_keep_power, int, 1);
MODULE_PARM_DESC(pm_keep_power, "1: PM keep power; 0: PM no power");
#endif
module_param(init_cfg, charp, 0);
MODULE_PARM_DESC(init_cfg, "BT init config file name");

module_param(drv_mode, int, 0);
MODULE_PARM_DESC(drv_mode, "Bit 0: BT/AMP/BLE; Bit 1: FM; Bit 2: NFC");
